@@ -313,8 +313,8 @@ class Medium(object):
                  epsilon_offdiag=Vector3(),
                  mu_diag=Vector3(1, 1, 1),
                  mu_offdiag=Vector3(),
-                 E_susceptibilities=[],
-                 H_susceptibilities=[],
+                 E_susceptibilities=None,
+                 H_susceptibilities=None,
                  E_chi2_diag=Vector3(),
                  E_chi3_diag=Vector3(),
                  H_chi2_diag=Vector3(),
@@ -425,8 +425,8 @@ class Medium(object):
         self.epsilon_offdiag = Vector3(*epsilon_offdiag)
         self.mu_diag = Vector3(*mu_diag)
         self.mu_offdiag = Vector3(*mu_offdiag)
-        self.E_susceptibilities = E_susceptibilities
-        self.H_susceptibilities = H_susceptibilities
+        self.E_susceptibilities = E_susceptibilities if E_susceptibilities else []
+        self.H_susceptibilities = H_susceptibilities if H_susceptibilities else []
         self.E_chi2_diag = Vector3(chi2, chi2, chi2) if chi2 else Vector3(*E_chi2_diag)
         self.E_chi3_diag = Vector3(chi3, chi3, chi3) if chi3 else Vector3(*E_chi3_diag)
         self.H_chi2_diag = Vector3(*H_chi2_diag)
@@ -517,13 +517,81 @@ class Medium(object):
         # Account for conductivity term (only multiply if nonzero to avoid unnecessary complex numbers)
         conductivity = np.expand_dims(Matrix(diag=conductivity_diag,offdiag=conductivity_offdiag),axis=0)
         if np.count_nonzero(conductivity) > 0:
-            epsmu = (1 + 1j/freqs * conductivity) * epsmu
+            epsmu = (1 + 1j/(2*np.pi*freqs) * conductivity) * epsmu
 
         # Convert list matrix to 3D numpy array size [freqs,3,3]
         return np.squeeze(epsmu)
 
 class MaterialGrid(object):
-    def __init__(self,grid_size,medium1,medium2,design_parameters=None,grid_type="U_DEFAULT"):
+    """
+    This class is used to specify materials on a rectilinear grid. A class object is passed
+    as the `material` argument of a [`Block`](#block) geometric object or the `default_material`
+    argument of the [`Simulation`](#Simulation) constructor (similar to a [material function](#medium)).
+    """
+    def check_weights(self, w):
+        if np.amin(w) < 0. or np.amax(w) > 1.:
+            warnings.warn('The weights parameter of MaterialGrid must be in the range [0,1].')
+            return np.clip(w, 0., 1.)
+        else:
+            return w
+
+    def __init__(self,
+                 grid_size,
+                 medium1,
+                 medium2,
+                 weights=None,
+                 grid_type="U_DEFAULT",
+                 do_averaging=True,
+                 beta=0,
+                 eta=0.5,
+                 damping=0):
+        """
+        Creates a `MaterialGrid` object.
+
+        The input are two materials `medium1` and `medium2` along with a weight function $u(x)$ which
+        is defined on a rectilinear grid by the NumPy array `weights` of size `grid_size` (a 3-tuple or
+        `Vector3` of integers $N_x$,$N_y$,$N_z$). The resolution of the grid may be nonuniform depending
+        on the `size` property of the `Block` object as shown in the following example for a 2d `MaterialGrid`
+        with $N_x=5$ and $N_y=4$. $N_z=0$ implies that the `MaterialGrid` is extruded in the $z$ direction.
+        The grid points are defined at the corners of the voxels.
+
+        ![](images/material_grid.png)
+
+        Elements of the `weights` array must be in the range [0,1] where 0 is `medium1` and 1 is `medium2`.
+        The `weights` array is used to define a linear interpolation from `medium1` to `medium2`.
+        Two material types are supported: (1) frequency-independent isotropic $\\varepsilon$ (`epsilon_diag`
+        and `epsilon_offdiag` are interpolated) and (2) `LorentzianSusceptibility` (`sigma` and `sigma_offdiag`
+        are interpolated). `medium1` and `medium2` must both be the same type. The materials are
+        [bilinearly interpolated](https://en.wikipedia.org/wiki/Bilinear_interpolation) from the rectilinear
+        grid to Meep's [Yee grid](Yee_Lattice.md).
+
+        For improving accuracy, [subpixel smoothing](Subpixel_Smoothing.md) can be enabled by specifying
+        `do_averaging=True`. If you want to use a material grid to define a (nearly) discontinuous,
+        piecewise-constant material that is *either* `medium1` or `medium2` almost everywhere, you can
+        optionally enable a (smoothed) *projection* feature by setting the parameter `beta` to a
+        positive value. When the projection feature is enabled, the weights $u(x)$ can be thought of as a
+        [level-set function](https://en.wikipedia.org/wiki/Level-set_method) defining an interface at
+        $u(x)=\\eta$ with a smoothing factor $\\beta$ where $\\beta=+\\infty$ gives an unsmoothed,
+        discontinuous interface. The projection operator is $(\\tanh(\\beta\\times\\eta)
+        +\\tanh(\\beta\\times(u-\\eta)))/(\\tanh(\\beta\\times\\eta)+\\tanh(\\beta\\times(1-\\eta)))$
+        involving the parameters `beta` ($\\beta$: bias or "smoothness" of the turn on) and `eta`
+        ($\\eta$: offset for erosion/dilation). The level set provides a general approach for defining
+        a *discontinuous* function from otherwise continuously varying (via the bilinear interpolation)
+        grid values. Subpixel smoothing is fast and accurate because it exploits an analytic formulation
+        for level-set functions.
+
+        A nonzero `damping` term creates an artificial conductivity $\\sigma = u(1-u)*$`damping`, which acts as
+        dissipation loss that penalizes intermediate pixel values of non-binarized structures. The value of
+        `damping` should be proportional to $2\\pi$ times the typical frequency of the problem.
+
+        It is possible to overlap any number of different `MaterialGrid`s. This can be useful for defining
+        grids which are symmetric (e.g., mirror, rotation). One way to set this up is by overlapping a
+        given `MaterialGrid` object with a symmetrized copy of itself. In the case of spatially overlapping
+        `MaterialGrid` objects (with no intervening objects), any overlapping points are computed using the
+        method `grid_type` which is one of `"U_MIN"` (minimum of the overlapping grid values), `"U_PROD"`
+        (product), `"U_MEAN"` (mean), `"U_DEFAULT"` (topmost material grid). In general, these `"U_*"` options
+        allow you to combine any material grids that overlap in space with no intervening objects.
+        """
         self.grid_size = mp.Vector3(*grid_size)
         self.medium1 = medium1
         self.medium2 = medium2
@@ -536,44 +604,51 @@ class MaterialGrid(object):
         if isclose(self.grid_size.z,0):
             self.grid_size.z = 1
         self.num_params=int(self.grid_size.x*self.grid_size.y*self.grid_size.z)
-
-        if design_parameters is None:
-            self.design_parameters = np.zeros((self.num_params,))
-        elif design_parameters.size != self.num_params:
-            raise ValueError("design_parameters of shape {} do not match user specified grid dimension: {}".format(design_parameters.size,self.grid_size))
+        self.do_averaging = do_averaging
+        self.beta = beta
+        self.eta = eta
+        self.damping = damping
+        if weights is None:
+            self.weights = np.zeros((self.num_params,))
+        elif weights.size != self.num_params:
+            raise ValueError("weights of shape {} do not match user specified grid dimension: {}".format(weights.size,self.grid_size))
         else:
-            self.design_parameters = design_parameters.flatten().astype(np.float64)
+            self.weights = self.check_weights(weights).flatten().astype(np.float64)
 
         grid_type_dict = {
             "U_MIN":0,
             "U_PROD":1,
-            "U_SUM":2,
+            "U_MEAN":2,
             "U_DEFAULT":3
         }
         if grid_type not in grid_type_dict:
-            raise ValueError("Invalid grid_type: {}. Must be either U_MIN, U_PROD, U_SUM, or U_DEFAULT".format(grid_type_dict))
+            raise ValueError("Invalid grid_type: {}. Must be either U_MIN, U_PROD, U_MEAN, or U_DEFAULT".format(grid_type_dict))
         self.grid_type = grid_type_dict[grid_type]
 
         self.swigobj = None
-    def update_parameters(self,x):
+
+    def update_weights(self, x):
+        """
+        Reset the `weights` to `x`.
+        """
         if x.size != self.num_params:
-            raise ValueError("design_parameters of shape {} do not match user specified grid dimension: {}".format(self.design_parameters.size,self.grid_size))
-        self.design_parameters[:]=x.flatten().astype(np.float64)
+            raise ValueError("weights of shape {} do not match user specified grid dimension: {}".format(self.weights.size,self.grid_size))
+        self.weights[:] = self.check_weights(x).flatten().astype(np.float64)
 
 class Susceptibility(object):
     """
     Parent class for various dispersive susceptibility terms, parameterized by an
-    anisotropic amplitude σ. See [Material Dispersion](Materials.md#material-dispersion).
+    anisotropic amplitude $\\sigma$. See [Material Dispersion](Materials.md#material-dispersion).
     """
     def __init__(self, sigma_diag=Vector3(), sigma_offdiag=Vector3(), sigma=None):
         """
-        + **`sigma` [`number`]** — The scale factor σ.
+        + **`sigma` [`number`]** — The scale factor $\\sigma$.
 
-        You can also specify an anisotropic σ tensor by using the property `sigma_diag`
-        which takes three numbers or a `Vector3` to give the σ$_n$ tensor diagonal, and
+        You can also specify an anisotropic $\\sigma$ tensor by using the property `sigma_diag`
+        which takes three numbers or a `Vector3` to give the $\\sigma_n$ tensor diagonal, and
         `sigma_offdiag` which specifies the offdiagonal elements (defaults to 0). That is,
         `sigma_diag=mp.Vector3(a, b, c)` and `sigma_offdiag=mp.Vector3(u, v, w)`
-        corresponds to a σ tensor
+        corresponds to a $\\sigma$ tensor
 
         \\begin{pmatrix} a & u & v \\\\ u & b & w \\\\ v & w & c \\end{pmatrix}
         """
@@ -591,13 +666,13 @@ class LorentzianSusceptibility(Susceptibility):
     """
     Specifies a single dispersive susceptibility of Lorentzian (damped harmonic
     oscillator) form. See [Material Dispersion](Materials.md#material-dispersion), with
-    the parameters (in addition to σ):
+    the parameters (in addition to $\\sigma$):
     """
     def __init__(self, frequency=0.0, gamma=0.0, **kwargs):
         """
         + **`frequency` [`number`]** — The resonance frequency $f_n = \\omega_n / 2\\pi$.
 
-        + **`gamma` [`number`]** — The resonance loss rate $γ_n / 2\\pi$.
+        + **`gamma` [`number`]** — The resonance loss rate $\\gamma_n / 2\\pi$.
 
         Note: multiple objects with identical values for the `frequency` and `gamma` but
         different `sigma` will appear as a *single* Lorentzian susceptibility term in the
@@ -618,14 +693,14 @@ class LorentzianSusceptibility(Susceptibility):
 class DrudeSusceptibility(Susceptibility):
     """
     Specifies a single dispersive susceptibility of Drude form. See [Material
-    Dispersion](Materials.md#material-dispersion), with the parameters (in addition to σ):
+    Dispersion](Materials.md#material-dispersion), with the parameters (in addition to $\\sigma$):
     """
     def __init__(self, frequency=0.0, gamma=0.0, **kwargs):
         """
         + **`frequency` [`number`]** — The frequency scale factor $f_n = \\omega_n / 2\\pi$
-          which multiplies σ (not a resonance frequency).
+          which multiplies $\\sigma$ (not a resonance frequency).
 
-        + **`gamma` [`number`]** — The loss rate $γ_n / 2\\pi$.
+        + **`gamma` [`number`]** — The loss rate $\\gamma_n / 2\\pi$.
         """
         super(DrudeSusceptibility, self).__init__(**kwargs)
         self.frequency = frequency
@@ -785,10 +860,10 @@ class MultilevelAtom(Susceptibility):
     atomic level. See [Materials/Saturable Gain and
     Absorption](Materials.md#saturable-gain-and-absorption).
     """
-    def __init__(self, initial_populations=[], transitions=[], **kwargs):
+    def __init__(self, initial_populations=None, transitions=None, **kwargs):
         super(MultilevelAtom, self).__init__(**kwargs)
-        self.initial_populations = initial_populations
-        self.transitions = transitions
+        self.initial_populations = initial_populations if initial_populations else []
+        self.transitions = transitions if transitions else []
 
 
 class Transition(object):
@@ -1108,6 +1183,9 @@ class Prism(GeometricObject):
           [centroid](https://en.wikipedia.org/wiki/Centroid) of all the vertices (top and
           bottom) of the resulting 3d prism so that the coordinates of the `vertices` are
           shifted accordingly.
+
+        + **`sidewall_angle` [`number`]** — The sidewall angle of the prism in units of
+          radians. Default is 0.
         """
 
         centroid = sum(vertices, Vector3(0)) * (1.0 / len(vertices)) # centroid of floor polygon
