@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology.
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,8 +26,6 @@
 
 using namespace std;
 
-typedef complex<double> cdouble;
-
 namespace meep {
 
 std::vector<double> linspace(double freq_min, double freq_max, size_t Nfreq) {
@@ -52,14 +50,15 @@ struct dft_chunk_data { // for passing to field::loop_in_chunks as void*
   bool sqrt_dV_and_interp_weights;
   bool empty_dim[5];
   dft_chunk *dft_chunks;
+  int decimation_factor;
 };
 
 dft_chunk::dft_chunk(fields_chunk *fc_, ivec is_, ivec ie_, vec s0_, vec s1_, vec e0_, vec e1_,
                      double dV0_, double dV1_, component c_, bool use_centered_grid,
-                     cdouble phase_factor, ivec shift_, const symmetry &S_, int sn_,
+                     complex<double> phase_factor, ivec shift_, const symmetry &S_, int sn_,
                      const void *data_) {
   dft_chunk_data *data = (dft_chunk_data *)data_;
-  if (!fc_->f[c_][0]) abort("invalid fields_chunk/component combination in dft_chunk");
+  if (!fc_->f[c_][0]) meep::abort("invalid fields_chunk/component combination in dft_chunk");
 
   fc = fc_;
   is = is_;
@@ -94,6 +93,7 @@ dft_chunk::dft_chunk(fields_chunk *fc_, ivec is_, ivec ie_, vec s0_, vec s1_, ve
   S = S_;
   sn = sn_;
   vc = data->vc;
+  decimation_factor = data->decimation_factor;
 
   const int Nomega = data->omega.size();
   omega = data->omega;
@@ -158,27 +158,47 @@ static void add_dft_chunkloop(fields_chunk *fc, int ichunk, component cgrid, ive
 dft_chunk *fields::add_dft(component c, const volume &where, const double *freq, size_t Nfreq,
                            bool include_dV_and_interp_weights, complex<double> stored_weight,
                            dft_chunk *chunk_next, bool sqrt_dV_and_interp_weights,
-                           complex<double> extra_weight, bool use_centered_grid, int vc) {
+                           complex<double> extra_weight, bool use_centered_grid,
+                           int vc, int decimation_factor) {
   if (coordinate_mismatch(gv.dim, c)) return NULL;
 
   /* If you call add_dft before adding sources, it will do nothing
      since no fields will be found.   This is almost certainly not
      what the user wants. */
   if (!components_allocated)
-    abort("allocate field components (by adding sources) before adding dft objects");
+    meep::abort("allocate field components (by adding sources) before adding dft objects");
   if (!include_dV_and_interp_weights && sqrt_dV_and_interp_weights)
-    abort("include_dV_and_interp_weights must be true for sqrt_dV_and_interp_weights=true in "
+    meep::abort("include_dV_and_interp_weights must be true for sqrt_dV_and_interp_weights=true in "
           "add_dft");
 
   dft_chunk_data data;
   data.c = c;
   data.vc = vc;
+
+  if (decimation_factor == 0) {
+    double src_freq_max = 0;
+    for (src_time *s = sources; s; s = s->next) {
+      if (s->get_fwidth() == 0)
+        decimation_factor = 1;
+      else
+        src_freq_max = std::max(src_freq_max, std::abs(s->frequency().real())+0.5*s->get_fwidth());
+    }
+    double freq_max = 0;
+    for (size_t i = 0; i < Nfreq; ++i)
+      freq_max = std::max(freq_max, std::abs(freq[i]));
+    if ((freq_max > 0) && (src_freq_max > 0))
+      decimation_factor = std::max(1, int(std::floor(1/(dt*(freq_max + src_freq_max)))));
+    else
+      decimation_factor = 1;
+  }
+  data.decimation_factor = decimation_factor;
+
   data.omega.resize(Nfreq);
   for (size_t i = 0; i < Nfreq; ++i)
     data.omega[i] = 2 * pi * freq[i];
   data.stored_weight = stored_weight;
   data.extra_weight = extra_weight;
-  data.dt_factor = dt / sqrt(2.0 * pi);
+  data.dt_factor = dt / sqrt(2.0 * pi) * decimation_factor;
   data.include_dV_and_interp_weights = include_dV_and_interp_weights;
   data.sqrt_dV_and_interp_weights = sqrt_dV_and_interp_weights;
   data.empty_dim[0] = data.empty_dim[1] = data.empty_dim[2] = data.empty_dim[3] =
@@ -190,12 +210,12 @@ dft_chunk *fields::add_dft(component c, const volume &where, const double *freq,
   return data.dft_chunks;
 }
 
-dft_chunk *fields::add_dft(const volume_list *where, const std::vector<double> freq,
+dft_chunk *fields::add_dft(const volume_list *where, const std::vector<double> &freq,
                            bool include_dV_and_interp_weights) {
   dft_chunk *chunks = 0;
   while (where) {
-    if (is_derived(where->c)) abort("derived_component invalid for dft");
-    cdouble stored_weight = where->weight;
+    if (is_derived(where->c)) meep::abort("derived_component invalid for dft");
+    complex<double> stored_weight = where->weight;
     chunks = add_dft(component(where->c), where->v, freq, include_dV_and_interp_weights,
                      stored_weight, chunks);
     where = where->next;
@@ -206,14 +226,16 @@ dft_chunk *fields::add_dft(const volume_list *where, const std::vector<double> f
 void fields::update_dfts() {
   am_now_working_on(FourierTransforming);
   for (int i = 0; i < num_chunks; i++)
-    if (chunks[i]->is_mine()) chunks[i]->update_dfts(time(), time() - 0.5 * dt);
+    if (chunks[i]->is_mine()) chunks[i]->update_dfts(time(), time() - 0.5 * dt, t);
   finished_working();
 }
 
-void fields_chunk::update_dfts(double timeE, double timeH) {
+void fields_chunk::update_dfts(double timeE, double timeH, int current_step) {
   if (doing_solve_cw) return;
   for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_chunk) {
-    cur->update_dft(is_magnetic(cur->c) ? timeH : timeE);
+    if ((current_step % cur->get_decimation_factor()) == 0) {
+      cur->update_dft(is_magnetic(cur->c) ? timeH : timeE);
+    }
   }
 }
 
@@ -226,8 +248,8 @@ void dft_chunk::update_dft(double time) {
 
   int numcmp = fc->f[c][1] ? 2 : 1;
 
-  size_t idx_dft = 0;
-  LOOP_OVER_IVECS(fc->gv, is, ie, idx) {
+  PLOOP_OVER_IVECS(fc->gv, is, ie, idx) {
+    size_t idx_dft = IVEC_LOOP_COUNTER;
     double w;
     if (include_dV_and_interp_weights) {
       w = IVEC_LOOP_WEIGHT(s0, s1, e0, e1, dV0 + dV1 * loop_i2);
@@ -235,7 +257,7 @@ void dft_chunk::update_dft(double time) {
     }
     else
       w = 1.0;
-    double f[2]; // real/imag field value at epsilon point
+    realnum f[2]; // real/imag field value at epsilon point
     if (avg2)
       for (int cmp = 0; cmp < numcmp; ++cmp)
         f[cmp] = (w * 0.25) * (fc->f[c][cmp][idx] + fc->f[c][cmp][idx + avg1] +
@@ -255,10 +277,83 @@ void dft_chunk::update_dft(double time) {
     else {
       realnum fr = f[0];
       for (int i = 0; i < Nomega; ++i)
-        dft[Nomega * idx_dft + i] += dft_phase[i] * fr;
+        dft[Nomega * idx_dft + i] += std::complex<realnum>{fr * dft_phase[i].real(), fr * dft_phase[i].imag()};
     }
+  }
+}
+
+/* Return the L2 norm of the DFTs themselves.  This is useful
+   to check whether the simulation is finished (whether all relevant fields have decayed).
+   (Collective operation.) */
+double fields::dft_norm() {
+  am_now_working_on(Other);
+  double sum = 0.0;
+  for (int i = 0; i < num_chunks; i++)
+    if (chunks[i]->is_mine()) sum += chunks[i]->dft_norm2();
+  finished_working();
+  return std::sqrt(sum_to_all(sum));
+}
+
+double fields_chunk::dft_norm2() const {
+  double sum = 0.0;
+  for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_chunk)
+    sum += cur->norm2();
+  return sum;
+}
+
+static double sqr(std::complex<realnum> x) { return (x*std::conj(x)).real(); }
+
+double dft_chunk::norm2() const {
+  if (!fc->f[c][0]) return 0.0;
+  double sum = 0.0;
+  size_t idx_dft = 0;
+  const int Nomega = omega.size();
+  LOOP_OVER_IVECS(fc->gv, is, ie, idx) {
+    for (int i = 0; i < Nomega; ++i)
+        sum += sqr(dft[Nomega * idx_dft + i]);
     idx_dft++;
   }
+  return sum;
+}
+
+// return the maximum decimation factor across
+// all dft regions
+int fields::max_decimation() const {
+  int maxdec = 1;
+  for (int i = 0; i < num_chunks; i++)
+    if (chunks[i]->is_mine())
+      maxdec = std::max(maxdec, chunks[i]->max_decimation());
+  return max_to_all(maxdec);
+}
+
+int fields_chunk::max_decimation() const {
+  int maxdec = std::numeric_limits<int>::min();
+  for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_chunk)
+    maxdec = std::max(maxdec, cur->get_decimation_factor());
+  return maxdec;
+}
+
+// return the maximum abs(freq) over all DFT chunks
+double fields::dft_maxfreq() const {
+  double maxfreq = 0;
+  for (int i = 0; i < num_chunks; i++)
+    if (chunks[i]->is_mine())
+      maxfreq = std::max(maxfreq, chunks[i]->dft_maxfreq());
+  return max_to_all(maxfreq);
+}
+
+double fields_chunk::dft_maxfreq() const {
+  double maxomega = 0;
+  for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_chunk)
+    maxomega = std::max(maxomega, cur->maxomega());
+  return maxomega / (2*meep::pi);
+}
+
+double dft_chunk::maxomega() const {
+  double maxomega = 0;
+  for (const auto& o : omega)
+    maxomega = std::max(maxomega, std::abs(o));
+  return maxomega;
 }
 
 void dft_chunk::scale_dft(complex<double> scale) {
@@ -269,29 +364,44 @@ void dft_chunk::scale_dft(complex<double> scale) {
 
 void dft_chunk::operator-=(const dft_chunk &chunk) {
   if (c != chunk.c || N * omega.size() != chunk.N * chunk.omega.size())
-    abort("Mismatched chunks in dft_chunk::operator-=");
+    meep::abort("Mismatched chunks in dft_chunk::operator-=");
 
   for (size_t i = 0; i < N * omega.size(); ++i)
     dft[i] -= chunk.dft[i];
 
   if (next_in_dft) {
-    if (!chunk.next_in_dft) abort("Mismatched chunk lists in dft_chunk::operator-=");
+    if (!chunk.next_in_dft) meep::abort("Mismatched chunk lists in dft_chunk::operator-=");
     *next_in_dft -= *chunk.next_in_dft;
   }
 }
 
-size_t dft_chunks_Ntotal(dft_chunk *dft_chunks, size_t *my_start) {
-  size_t n = 0;
+size_t my_dft_chunks_Ntotal(dft_chunk *dft_chunks, size_t *my_start) {
+  // When writing to a sharded file, we write out only the chunks we own.
+size_t n = 0;
   for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_dft)
     n += cur->N * cur->omega.size() * 2;
+
+  *my_start = 0;
+  return n;
+}
+
+size_t dft_chunks_Ntotal(dft_chunk *dft_chunks, size_t *my_start) {
+  // If writing to a single parallel file, we are compute our chunks offset
+  // into the single-parallel-file that has all the data.
+  size_t n = my_dft_chunks_Ntotal(dft_chunks, my_start);
   *my_start = partial_sum_to_all(n) - n; // sum(n) for processes before this
   return sum_to_all(n);
 }
 
+size_t dft_chunks_Ntotal(dft_chunk *dft_chunks, size_t *my_start, bool single_parallel_file) {
+  return single_parallel_file ? dft_chunks_Ntotal(dft_chunks, my_start) :
+                                my_dft_chunks_Ntotal(dft_chunks, my_start);
+}
+
 // Note: the file must have been created in parallel mode, typically via fields::open_h5file.
-void save_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const char *dprefix) {
+void save_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const char *dprefix, bool single_parallel_file) {
   size_t istart;
-  size_t n = dft_chunks_Ntotal(dft_chunks, &istart);
+  size_t n = dft_chunks_Ntotal(dft_chunks, &istart, single_parallel_file);
 
   char dataname[1024];
   snprintf(dataname, 1024,
@@ -308,13 +418,13 @@ void save_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const
   file->done_writing_chunks();
 }
 
-void save_dft_hdf5(dft_chunk *dft_chunks, component c, h5file *file, const char *dprefix) {
-  save_dft_hdf5(dft_chunks, component_name(c), file, dprefix);
+void save_dft_hdf5(dft_chunk *dft_chunks, component c, h5file *file, const char *dprefix, bool single_parallel_file) {
+  save_dft_hdf5(dft_chunks, component_name(c), file, dprefix, single_parallel_file);
 }
 
-void load_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const char *dprefix) {
+void load_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const char *dprefix, bool single_parallel_file) {
   size_t istart;
-  size_t n = dft_chunks_Ntotal(dft_chunks, &istart);
+  size_t n = dft_chunks_Ntotal(dft_chunks, &istart, single_parallel_file);
 
   char dataname[1024];
   snprintf(dataname, 1024,
@@ -325,7 +435,7 @@ void load_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const
   size_t file_dims;
   file->read_size(dataname, &file_rank, &file_dims, 1);
   if (file_rank != 1 || file_dims != n)
-    abort("incorrect dataset size (%zd vs. %zd) in load_dft_hdf5 %s:%s", file_dims, n,
+    meep::abort("incorrect dataset size (%zd vs. %zd) in load_dft_hdf5 %s:%s", file_dims, n,
           file->file_name(), dataname);
 
   for (dft_chunk *cur = dft_chunks; cur; cur = cur->next_in_dft) {
@@ -335,8 +445,8 @@ void load_dft_hdf5(dft_chunk *dft_chunks, const char *name, h5file *file, const
   }
 }
 
-void load_dft_hdf5(dft_chunk *dft_chunks, component c, h5file *file, const char *dprefix) {
-  load_dft_hdf5(dft_chunks, component_name(c), file, dprefix);
+void load_dft_hdf5(dft_chunk *dft_chunks, component c, h5file *file, const char *dprefix, bool single_parallel_file) {
+  load_dft_hdf5(dft_chunks, component_name(c), file, dprefix, single_parallel_file);
 }
 
 dft_flux::dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_chunk *H_,
@@ -348,7 +458,7 @@ dft_flux::dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_
 }
 
 dft_flux::dft_flux(const component cE_, const component cH_, dft_chunk *E_, dft_chunk *H_,
-                   const std::vector<double> freq_, const volume &where_,
+                   const std::vector<double> &freq_, const volume &where_,
                    direction normal_direction_, bool use_symmetry_)
     : E(E_), H(H_), cE(cE_), cH(cH_), where(where_), normal_direction(normal_direction_),
       use_symmetry(use_symmetry_) {
@@ -420,7 +530,7 @@ void dft_flux::scale_dfts(complex<double> scale) {
 }
 
 dft_flux fields::add_dft_flux(const volume_list *where_, const double *freq, size_t Nfreq,
-                              bool use_symmetry, bool centered_grid) {
+                              bool use_symmetry, bool centered_grid, int decimation_factor) {
   if (!where_) // handle empty list of volumes
     return dft_flux(Ex, Hy, NULL, NULL, freq, Nfreq, v, NO_DIRECTION, use_symmetry);
 
@@ -438,7 +548,7 @@ dft_flux fields::add_dft_flux(const volume_list *where_, const double *freq, siz
   while (where) {
     derived_component c = derived_component(where->c);
     if (coordinate_mismatch(gv.dim, component_direction(c)))
-      abort("coordinate-type mismatch in add_dft_flux");
+      meep::abort("coordinate-type mismatch in add_dft_flux");
 
     switch (c) {
       case Sx: cE[0] = Ey, cE[1] = Ez, cH[0] = Hz, cH[1] = Hy; break;
@@ -451,13 +561,15 @@ dft_flux fields::add_dft_flux(const volume_list *where_, const double *freq, siz
         else
           cE[0] = Ex, cE[1] = Ey, cH[0] = Hy, cH[1] = Hx;
         break;
-      default: abort("invalid flux component!");
+      default: meep::abort("invalid flux component!");
     }
 
     for (int i = 0; i < 2; ++i) {
       E = add_dft(cE[i], where->v, freq, Nfreq, true,
-                  where->weight * double(1 - 2 * i), E, false, std::complex<double>(1.0,0), centered_grid);
-      H = add_dft(cH[i], where->v, freq, Nfreq, false, 1.0, H, false, std::complex<double>(1.0,0), centered_grid);
+                  where->weight * double(1 - 2 * i), E, false, std::complex<double>(1.0,0),
+                  centered_grid, 0, decimation_factor);
+      H = add_dft(cH[i], where->v, freq, Nfreq, false, 1.0, H, false, std::complex<double>(1.0,0),
+                  centered_grid, 0, decimation_factor);
     }
 
     where = where->next;
@@ -477,7 +589,7 @@ dft_energy::dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B
 }
 
 dft_energy::dft_energy(dft_chunk *E_, dft_chunk *H_, dft_chunk *D_, dft_chunk *B_,
-                       const std::vector<double> freq_, const volume &where_)
+                       const std::vector<double> &freq_, const volume &where_)
     : E(E_), H(H_), D(D_), B(B_), where(where_) {
   freq = freq_;
 }
@@ -541,7 +653,8 @@ double *dft_energy::total() {
   return F;
 }
 
-dft_energy fields::add_dft_energy(const volume_list *where_, const double *freq, size_t Nfreq) {
+dft_energy fields::add_dft_energy(const volume_list *where_, const double *freq, size_t Nfreq,
+                                  int decimation_factor) {
 
   if (!where_) // handle empty list of volumes
     return dft_energy(NULL, NULL, NULL, NULL, freq, Nfreq, v);
@@ -552,10 +665,14 @@ dft_energy fields::add_dft_energy(const volume_list *where_, const double *freq,
   volume_list *where_save = where;
   while (where) {
     LOOP_OVER_FIELD_DIRECTIONS(gv.dim, d) {
-      E = add_dft(direction_component(Ex, d), where->v, freq, Nfreq, true, 1.0, E);
-      D = add_dft(direction_component(Dx, d), where->v, freq, Nfreq, false, 1.0, D);
-      H = add_dft(direction_component(Hx, d), where->v, freq, Nfreq, true, 1.0, H);
-      B = add_dft(direction_component(Bx, d), where->v, freq, Nfreq, false, 1.0, B);
+      E = add_dft(direction_component(Ex, d), where->v, freq, Nfreq, true, 1.0, E,
+                  false, 1.0, true, 0, decimation_factor);
+      D = add_dft(direction_component(Dx, d), where->v, freq, Nfreq, false, 1.0, D,
+                  false, 1.0, true, 0, decimation_factor);
+      H = add_dft(direction_component(Hx, d), where->v, freq, Nfreq, true, 1.0, H,
+                  false, 1.0, true, 0, decimation_factor);
+      B = add_dft(direction_component(Bx, d), where->v, freq, Nfreq, false, 1.0, B,
+                  false, 1.0, true, 0, decimation_factor);
     }
     where = where->next;
   }
@@ -640,23 +757,24 @@ direction fields::normal_direction(const volume &where) const {
     if (d == NO_DIRECTION && gv.dim == D2 && beta != 0 && where_pad.in_direction(X) > 0 &&
         where_pad.in_direction(Y) > 0)
       d = Z;
-    if (d == NO_DIRECTION) abort("Could not determine normal direction for given grid_volume.");
+    if (d == NO_DIRECTION) meep::abort("Could not determine normal direction for given grid_volume.");
   }
   return d;
 }
 
 dft_flux fields::add_dft_flux(direction d, const volume &where, const double *freq, size_t Nfreq,
-                              bool use_symmetry, bool centered_grid) {
+                              bool use_symmetry, bool centered_grid, int decimation_factor) {
   if (d == NO_DIRECTION) d = normal_direction(where);
   volume_list vl(where, direction_component(Sx, d));
-  dft_flux flux = add_dft_flux(&vl, freq, Nfreq, use_symmetry, centered_grid);
+  dft_flux flux = add_dft_flux(&vl, freq, Nfreq, use_symmetry, centered_grid, decimation_factor);
   flux.normal_direction = d;
   return flux;
 }
 
 
-dft_flux fields::add_mode_monitor(direction d, const volume &where, const double *freq, size_t Nfreq, bool centered_grid) {
-  return add_dft_flux(d, where, freq, Nfreq, /*use_symmetry=*/false, centered_grid);
+dft_flux fields::add_mode_monitor(direction d, const volume &where, const double *freq,
+                                  size_t Nfreq, bool centered_grid, int decimation_factor) {
+  return add_dft_flux(d, where, freq, Nfreq, /*use_symmetry=*/false, centered_grid, decimation_factor);
 }
 
 dft_flux fields::add_dft_flux_box(const volume &where, double freq_min, double freq_max,
@@ -664,7 +782,7 @@ dft_flux fields::add_dft_flux_box(const volume &where, double freq_min, double f
   return add_dft_flux_box(where, meep::linspace(freq_min, freq_max, Nfreq));
 }
 
-dft_flux fields::add_dft_flux_box(const volume &where, const std::vector<double> freq) {
+dft_flux fields::add_dft_flux_box(const volume &where, const std::vector<double> &freq) {
   volume_list *faces = 0;
   LOOP_OVER_DIRECTIONS(where.dim, d) {
     if (where.in_direction(d) > 0) {
@@ -688,7 +806,7 @@ dft_flux fields::add_dft_flux_plane(const volume &where, double freq_min, double
   return add_dft_flux_plane(where, meep::linspace(freq_min, freq_max, Nfreq));
 }
 
-dft_flux fields::add_dft_flux_plane(const volume &where, const std::vector<double> freq) {
+dft_flux fields::add_dft_flux_plane(const volume &where, const std::vector<double> &freq) {
   return add_dft_flux(NO_DIRECTION, where, freq);
 }
 
@@ -699,7 +817,7 @@ dft_fields::dft_fields(dft_chunk *chunks_, double freq_min, double freq_max, int
   freq = meep::linspace(freq_min, freq_max, Nf);
 }
 
-dft_fields::dft_fields(dft_chunk *chunks_, const std::vector<double> freq_, const volume &where_)
+dft_fields::dft_fields(dft_chunk *chunks_, const std::vector<double> &freq_, const volume &where_)
     : where(where_) {
   chunks = chunks_;
   freq = freq_;
@@ -712,7 +830,7 @@ dft_fields::dft_fields(dft_chunk *chunks_, const double *freq_, size_t Nfreq, co
     freq[i] = freq_[i];
 }
 
-void dft_fields::scale_dfts(cdouble scale) { chunks->scale_dft(scale); }
+void dft_fields::scale_dfts(complex<double> scale) { chunks->scale_dft(scale); }
 
 void dft_fields::remove() {
   while (chunks) {
@@ -723,16 +841,17 @@ void dft_fields::remove() {
 }
 
 dft_fields fields::add_dft_fields(component *components, int num_components, const volume where,
-                                  const double *freq, size_t Nfreq, bool use_centered_grid) {
+                                  const double *freq, size_t Nfreq, bool use_centered_grid,
+                                  int decimation_factor) {
   bool include_dV_and_interp_weights = false;
   bool sqrt_dV_and_interp_weights = false; // default option from meep.hpp (expose to user?)
   std::complex<double> extra_weight = 1.0; // default option from meep.hpp (expose to user?)
-  cdouble stored_weight = 1.0;
-  dft_chunk *chunks = 0;
+  complex<double> stored_weight = 1.0;
+  dft_chunk *chunks = NULL;
   for (int nc = 0; nc < num_components; nc++)
-    chunks =
-        add_dft(components[nc], where, freq, Nfreq, include_dV_and_interp_weights, stored_weight,
-                chunks, sqrt_dV_and_interp_weights, extra_weight, use_centered_grid);
+    chunks = add_dft(components[nc], where, freq, Nfreq, include_dV_and_interp_weights,
+                     stored_weight, chunks, sqrt_dV_and_interp_weights, extra_weight,
+                     use_centered_grid, 0, decimation_factor);
 
   return dft_fields(chunks, freq, Nfreq, where);
 }
@@ -740,14 +859,14 @@ dft_fields fields::add_dft_fields(component *components, int num_components, con
 /***************************************************************/
 /* chunk-level processing for fields::process_dft_component.   */
 /***************************************************************/
-cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corner, ivec max_corner,
-                                         int num_freq, h5file *file, double *buffer, int reim,
-                                         cdouble *field_array, void *mode1_data, void *mode2_data,
-                                         int ic_conjugate, bool retain_interp_weights,
-                                         fields *parent) {
+complex<double> dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corner, ivec max_corner,
+                                                 int num_freq, h5file *file, realnum *buffer, int reim,
+                                                 complex<realnum> *field_array, void *mode1_data, void *mode2_data,
+                                                 int ic_conjugate, bool retain_interp_weights,
+                                                 fields *parent) {
 
-  if ((num_freq < 0) || (num_freq > omega.size()-1))
-    abort("process_dft_component: frequency index %d is outside the range of the frequency array of size %lu",num_freq,omega.size());
+  if ((num_freq < 0) || (num_freq > static_cast<int>(omega.size())-1))
+    meep::abort("process_dft_component: frequency index %d is outside the range of the frequency array of size %lu",num_freq,omega.size());
 
   /*****************************************************************/
   /* compute the size of the chunk we own and its strides etc.     */
@@ -768,7 +887,7 @@ cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corne
   for (int i = 0; i < rank; ++i) {
     direction d = ds[i];
     int isd = isS.in_direction(d), ied = ieS.in_direction(d);
-    start[i] = (min(isd, ied) - min_corner.in_direction(d)) / 2;
+    start[i] = (std::min(isd, ied) - min_corner.in_direction(d)) / 2;
     file_count[i] = abs(ied - isd) / 2 + 1;
     if (ied < isd) file_offset[permute.in_direction(d)] = file_count[i] - 1;
     array_count[i] = (max_corner.in_direction(d) - min_corner.in_direction(d)) / 2 + 1;
@@ -804,7 +923,7 @@ cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corne
   /***************************************************************/
   vec rshift(shift * (0.5 * fc->gv.inva));
   int chunk_idx = 0;
-  cdouble integral = 0.0;
+  complex<double> integral = 0.0;
   component c_conjugate = (component)(ic_conjugate >= 0 ? ic_conjugate : -ic_conjugate);
   LOOP_OVER_IVECS(fc->gv, is, ie, idx) {
     IVEC_LOOP_LOC(fc->gv, loc);
@@ -812,17 +931,17 @@ cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corne
     double w = IVEC_LOOP_WEIGHT(s0, s1, e0, e1, dV0 + dV1 * loop_i2);
     double interp_w = retain_interp_weights ? IVEC_LOOP_WEIGHT(s0i, s1i, e0i, e1i, 1.0) : 1.0;
 
-    cdouble dft_val =
+    complex<double> dft_val =
         (c_conjugate == NO_COMPONENT
              ? w
              : c_conjugate == Dielectric
                    ? parent->get_eps(loc)
                    : c_conjugate == Permeability
                          ? parent->get_mu(loc)
-                         : dft[omega.size() * (chunk_idx++) + num_freq] / stored_weight);
+                         : complex<double>(dft[omega.size() * (chunk_idx++) + num_freq]) / stored_weight);
     if (include_dV_and_interp_weights) dft_val /= (sqrt_dV_and_interp_weights ? sqrt(w) : w);
 
-    cdouble mode1val = 0.0, mode2val = 0.0;
+    complex<double> mode1val = 0.0, mode2val = 0.0;
     if (mode1_data) mode1val = eigenmode_amplitude(mode1_data, loc, S.transform(c_conjugate, sn));
     if (mode2_data) mode2val = eigenmode_amplitude(mode2_data, loc, S.transform(c, sn));
 
@@ -833,7 +952,7 @@ cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corne
 
       dft_val *= interp_w;
 
-      cdouble val = (mode1_data ? mode1val : dft_val);
+      complex<double> val = (mode1_data ? mode1val : dft_val);
       buffer[idx2] = reim ? imag(val) : real(val);
     }
     else if (field_array) {
@@ -906,11 +1025,11 @@ cdouble dft_chunk::process_dft_component(int rank, direction *ds, ivec min_corne
 /* if where is non-null, only field components inside *where   */
 /* are processed.                                              */
 /***************************************************************/
-cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists, int num_freq,
-                                      component c, const char *HDF5FileName, cdouble **pfield_array,
-                                      int *array_rank, size_t *array_dims, direction *array_dirs,
-                                      void *mode1_data, void *mode2_data, component c_conjugate,
-                                      bool *first_component, bool retain_interp_weights) {
+complex<double> fields::process_dft_component(dft_chunk **chunklists, int num_chunklists, int num_freq,
+                                              component c, const char *HDF5FileName, complex<realnum> **pfield_array,
+                                              int *array_rank, size_t *array_dims, direction *array_dirs,
+                                              void *mode1_data, void *mode2_data, component c_conjugate,
+                                              bool *first_component, bool retain_interp_weights) {
 
   /***************************************************************/
   /***************************************************************/
@@ -940,9 +1059,9 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
       LOOP_OVER_DIRECTIONS(chunk->fc->gv.dim, d) {
         this_bufsz *= (chunk->ie.in_direction(d) - chunk->is.in_direction(d)) / 2 + 1;
       }
-      bufsz = max(bufsz, this_bufsz);
+      bufsz = std::max(bufsz, this_bufsz);
     }
-  am_now_working_on(MpiTime);
+  am_now_working_on(MpiAllTime);
   max_corner = max_to_all(max_corner);
   min_corner = -max_to_all(-min_corner); // i.e., min_to_all
   finished_working();
@@ -955,7 +1074,7 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
   direction ds[3];
   size_t array_size = 1;
   LOOP_OVER_DIRECTIONS(gv.dim, d) {
-    if (rank >= 3) abort("too many dimensions in process_dft_component");
+    if (rank >= 3) meep::abort("too many dimensions in process_dft_component");
     size_t n = std::max(0, (max_corner.in_direction(d) - min_corner.in_direction(d)) / 2 + 1);
 
     if (n > 1) {
@@ -981,18 +1100,16 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
   /* like h5_output_data::buf in h5fields.cpp                    */
   /***************************************************************/
   realnum *buffer = 0;
-  cdouble *field_array = 0;
+  complex<realnum> *field_array = 0;
   int reim_max = 0;
   if (HDF5FileName) {
     buffer = new realnum[bufsz];
     reim_max = 1;
   }
   else if (pfield_array)
-    *pfield_array = field_array = (array_size ? new cdouble[array_size] : 0);
+    *pfield_array = field_array = (array_size ? new complex<realnum>[array_size] : 0);
 
-  bool append_data = false;
-  bool single_precision = false;
-  cdouble overlap = 0.0;
+  complex<double> overlap = 0.0;
   for (int reim = 0; reim <= reim_max; reim++) {
     h5file *file = 0;
     if (HDF5FileName) {
@@ -1000,7 +1117,8 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
       *first_component = false;
       char dataname[100];
       snprintf(dataname, 100, "%s_%i.%c", component_name(c), num_freq, reim ? 'i' : 'r');
-      file->create_or_extend_data(dataname, rank, dims, append_data, single_precision);
+      file->create_or_extend_data(dataname, rank, dims, false /* append_data */,
+                                  sizeof(realnum) == sizeof(float) /* single_precision */);
     }
 
     for (int ncl = 0; ncl < num_chunklists; ncl++)
@@ -1020,16 +1138,16 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
 /* repeatedly call sum_to_all to consolidate full field array  */
 /* on all cores                                                */
 /***************************************************************/
-#define BUFSIZE 1 << 16 // use 64k buffer
-      cdouble *buf = new cdouble[BUFSIZE];
+#define BUFSIZE 1 << 20  // use 1M element (16 MB) buffer
+      complex<realnum> *buf = new complex<realnum>[BUFSIZE];
       ptrdiff_t offset = 0;
       size_t remaining = array_size;
       while (remaining != 0) {
         size_t size = (remaining > BUFSIZE ? BUFSIZE : remaining);
-        am_now_working_on(MpiTime);
+        am_now_working_on(MpiAllTime);
         sum_to_all(field_array + offset, buf, size);
         finished_working();
-        memcpy(field_array + offset, buf, size * sizeof(cdouble));
+        memcpy(field_array + offset, buf, size * sizeof(complex<realnum>));
         remaining -= size;
         offset += size;
       }
@@ -1040,7 +1158,7 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
   if (HDF5FileName)
     delete[] buffer;
   else {
-    am_now_working_on(MpiTime);
+    am_now_working_on(MpiAllTime);
     overlap = sum_to_all(overlap);
     finished_working();
   }
@@ -1051,46 +1169,46 @@ cdouble fields::process_dft_component(dft_chunk **chunklists, int num_chunklists
 /***************************************************************/
 /* routines for fetching arrays of dft fields                  */
 /***************************************************************/
-cdouble *collapse_array(cdouble *array, int *rank, size_t dims[3], direction dirs[3], volume where);
+complex<realnum> *collapse_array(complex<realnum> *array, int *rank, size_t dims[3], direction dirs[3], volume where);
 
-cdouble *fields::get_dft_array(dft_flux flux, component c, int num_freq, int *rank,
-                               size_t dims[3]) {
+complex<realnum> *fields::get_dft_array(dft_flux flux, component c, int num_freq, int *rank,
+                                        size_t dims[3]) {
   dft_chunk *chunklists[2];
   chunklists[0] = flux.E;
   chunklists[1] = flux.H;
-  cdouble *array;
+  complex<realnum> *array;
   direction dirs[3];
   process_dft_component(chunklists, 2, num_freq, c, 0, &array, rank, dims, dirs);
   return collapse_array(array, rank, dims, dirs, flux.where);
 }
 
-cdouble *fields::get_dft_array(dft_force force, component c, int num_freq, int *rank,
-                               size_t dims[3]) {
+complex<realnum> *fields::get_dft_array(dft_force force, component c, int num_freq, int *rank,
+                                        size_t dims[3]) {
   dft_chunk *chunklists[3];
   chunklists[0] = force.offdiag1;
   chunklists[1] = force.offdiag2;
   chunklists[2] = force.diag;
-  cdouble *array;
+  complex<realnum> *array;
   direction dirs[3];
   process_dft_component(chunklists, 3, num_freq, c, 0, &array, rank, dims, dirs);
   return collapse_array(array, rank, dims, dirs, force.where);
 }
 
-cdouble *fields::get_dft_array(dft_near2far n2f, component c, int num_freq, int *rank,
-                               size_t dims[3]) {
+complex<realnum> *fields::get_dft_array(dft_near2far n2f, component c, int num_freq, int *rank,
+                                        size_t dims[3]) {
   dft_chunk *chunklists[1];
   chunklists[0] = n2f.F;
-  cdouble *array;
+  complex<realnum> *array;
   direction dirs[3];
   process_dft_component(chunklists, 1, num_freq, c, 0, &array, rank, dims, dirs);
   return collapse_array(array, rank, dims, dirs, n2f.where);
 }
 
-cdouble *fields::get_dft_array(dft_fields fdft, component c, int num_freq, int *rank,
-                               size_t dims[3]) {
+complex<realnum> *fields::get_dft_array(dft_fields fdft, component c, int num_freq, int *rank,
+                                        size_t dims[3]) {
   dft_chunk *chunklists[1];
   chunklists[0] = fdft.chunks;
-  cdouble *array;
+  complex<realnum> *array;
   direction dirs[3];
   process_dft_component(chunklists, 1, num_freq, c, 0, &array, rank, dims, dirs);
   return collapse_array(array, rank, dims, dirs, fdft.where);
@@ -1125,7 +1243,7 @@ void fields::output_dft_components(dft_chunk **chunklists, int num_chunklists, v
     snprintf(filename, 100, "%s%s", HDF5FileName, strstr("%.h5", HDF5FileName) ? "" : ".h5");
     file = new h5file(filename, h5file::WRITE, false /*parallel*/);
   }
-  am_now_working_on(MpiTime);
+  am_now_working_on(MpiAllTime);
   if (have_empty_dims) NumFreqs = max_to_all(NumFreqs); // subtle!
   finished_working();
 
@@ -1137,7 +1255,7 @@ void fields::output_dft_components(dft_chunk **chunklists, int num_chunklists, v
                               0, Ex, &first_component);
       }
       else {
-        cdouble *array = 0;
+        complex<realnum> *array = 0;
         int rank;
         size_t dims[3];
         direction dirs[3];
@@ -1145,18 +1263,17 @@ void fields::output_dft_components(dft_chunk **chunklists, int num_chunklists, v
                               dirs);
         if (rank > 0 && am_master()) {
           array = collapse_array(array, &rank, dims, dirs, dft_volume);
-          if (rank == 0) abort("%s:%i: internal error", __FILE__, __LINE__);
+          if (rank == 0) meep::abort("%s:%i: internal error", __FILE__, __LINE__);
           size_t array_size = dims[0] * (rank >= 2 ? dims[1] * (rank == 3 ? dims[2] : 1) : 1);
           double *real_array = new double[array_size];
-          if (!real_array) abort("%s:%i:out of memory(%lu)", __FILE__, __LINE__, array_size);
+          if (!real_array) meep::abort("%s:%i:out of memory(%lu)", __FILE__, __LINE__, array_size);
           for (int reim = 0; reim < 2; reim++) {
             for (size_t n = 0; n < array_size; n++)
               real_array[n] = (reim == 0 ? real(array[n]) : imag(array[n]));
             char dataname[100], filename[100];
             snprintf(dataname, 100, "%s_%i.%c", component_name(c), num_freq, reim ? 'i' : 'r');
             snprintf(filename, 100, "%s%s", HDF5FileName, strstr(".h5", HDF5FileName) ? "" : ".h5");
-            bool single_precision = false;
-            file->write(dataname, rank, dims, real_array, single_precision);
+            file->write(dataname, rank, dims, real_array, false /* single_precision */);
           }
           delete[] real_array;
         }
@@ -1197,7 +1314,7 @@ void fields::output_dft(dft_fields fdft, const char *HDF5FileName) {
 /***************************************************************/
 /***************************************************************/
 void fields::get_overlap(void *mode1_data, void *mode2_data, dft_flux flux, int num_freq,
-                         cdouble overlaps[2]) {
+                         complex<double> overlaps[2]) {
   component cE[2], cH[2];
   switch (flux.normal_direction) {
     case X:
@@ -1230,20 +1347,20 @@ void fields::get_overlap(void *mode1_data, void *mode2_data, dft_flux flux, int
       else
         cE[0] = Ex, cE[1] = Ey, cH[0] = Hy, cH[1] = Hx;
       break;
-    default: abort("invalid normal_direction in get_overlap");
+    default: meep::abort("invalid normal_direction in get_overlap");
   };
 
   dft_chunk *chunklists[2];
   chunklists[0] = flux.E;
   chunklists[1] = flux.H;
-  cdouble ExHy = process_dft_component(chunklists, 2, num_freq, cE[0], 0, 0, 0, 0, 0, mode1_data,
-                                       mode2_data, cH[0]);
-  cdouble EyHx = process_dft_component(chunklists, 2, num_freq, cE[1], 0, 0, 0, 0, 0, mode1_data,
-                                       mode2_data, cH[1]);
-  cdouble HyEx = process_dft_component(chunklists, 2, num_freq, cH[0], 0, 0, 0, 0, 0, mode1_data,
-                                       mode2_data, cE[0]);
-  cdouble HxEy = process_dft_component(chunklists, 2, num_freq, cH[1], 0, 0, 0, 0, 0, mode1_data,
-                                       mode2_data, cE[1]);
+  complex<double> ExHy = process_dft_component(chunklists, 2, num_freq, cE[0], 0, 0, 0, 0, 0, mode1_data,
+                                               mode2_data, cH[0]);
+  complex<double> EyHx = process_dft_component(chunklists, 2, num_freq, cE[1], 0, 0, 0, 0, 0, mode1_data,
+                                               mode2_data, cH[1]);
+  complex<double> HyEx = process_dft_component(chunklists, 2, num_freq, cH[0], 0, 0, 0, 0, 0, mode1_data,
+                                               mode2_data, cE[0]);
+  complex<double> HxEy = process_dft_component(chunklists, 2, num_freq, cH[1], 0, 0, 0, 0, 0, mode1_data,
+                                               mode2_data, cE[1]);
   overlaps[0] = ExHy - EyHx;
   overlaps[1] = HyEx - HxEy;
 }
